# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.lock_db"""

import logging
import threading
import time

from sonicprobe.libs import anysql

from .abstract.db import XtrargonDBBase

from .common import STATUS_STOPPED


LOG                         = logging.getLogger('xtrargon_ng.classes.lock_db')

DEFAULT_DB_SLEEP_DURATION   = 0.5
DEFAULT_SAFE_SLAVE_TIMEOUT  = 300

DB_OPEN_TEMP_TABLES_WAIT    = 3
DB_LOCK_WAIT_TIMEOUT        = 31536000

DB_READ_ONLY_OPTS           = ('master',
                               'master-only',
                               'slave',
                               'slave-only')


class BackupsLockDB(threading.Thread, XtrargonDBBase):
    def __init__(self, module, config = None):
        threading.Thread.__init__(self)
        XtrargonDBBase.__init__(self)

        self.name                = 'lockdb'
        self._module             = module
        self._config             = config or {}
        self._lock_db            = module.lock_db
        self._db_locked          = module.db_locked
        self.init_db(module)

        self._killed             = False

        self._sleep_duration     = float(DEFAULT_DB_SLEEP_DURATION)
        self._sql_read_only      = False
        self._sql_thread_run     = False
        self._no_lock            = bool(self._config.get('no_lock', False))
        self._read_only          = None
        self._can_read_only      = None
        self._safe_slave         = bool(self._config.get('safe_slave', True))
        self._safe_slave_timeout = int(self._config.get('safe_slave_timeout', DEFAULT_SAFE_SLAVE_TIMEOUT))
        if self._safe_slave_timeout < 1:
            self._safe_slave_timeout = DEFAULT_SAFE_SLAVE_TIMEOUT

    def _stop(self):
        if self._dbcursor:
            LOG.info("DB UNLOCK TABLES")

            if self._can_read_only and not self._sql_read_only:
                self._dbcursor.query("SET GLOBAL read_only = OFF")

            self._dbcursor.query("UNLOCK TABLES")

            if self._sql_thread_run:
                self._dbcursor.query("START SLAVE SQL_THREAD")

        self.close_connection()

    def _get_config_read_only(self):
        if self._config.get('read_only') is None:
            return 'slave-only'

        if self._config.get('read_only') in DB_READ_ONLY_OPTS:
            return self._config.get('read_only')

        return bool(self._config.get('read_only'))

    def _check_read_only(self):
        if isinstance(self._read_only, bool):
            return self._read_only

        if self._read_only == 'slave':
            return self.check_is_slave()

        if self._read_only == 'master':
            return self.check_is_master()

        if self._read_only == 'slave-only':
            return self.check_is_slave() and not self.check_is_master()

        if self._read_only == 'master-only':
            return not self.check_is_master() and not self.check_is_slave()

        return False

    def _lock_tables_for_backup(self):
        if self._get_sql_variable('have_backup_locks') == 'YES':
            LOG.info("LOCK TABLES FOR BACKUP")
            self._dbcursor.query("LOCK TABLES FOR BACKUP")

            if self._can_read_only and not self._sql_read_only:
                self._dbcursor.query("SET GLOBAL read_only = ON")

            self._dbcursor.query("FLUSH LOGS")

            return True

        return False

    def _get_slave_coordinates(self):
        slave_status = self.get_slave_status()
        if not slave_status:
            return None

        return (slave_status['Relay_Master_Log_File'],
                slave_status['Exec_Master_Log_Pos'])

    def _get_slave_open_temp_tables(self):
        open_temp_tables = self._get_sql_variable('Slave_open_temp_tables')
        if not open_temp_tables:
            return 0

        return int(open_temp_tables)

    def _check_slave_open_temp_tables(self, start_time):
        prev_slave_coordinates = None
        curr_slave_coordinates = self._get_slave_coordinates()
        open_temp_tables       = None

        while True:
            if self._killed:
                raise GeneratorExit()

            open_temp_tables = self._get_slave_open_temp_tables()
            if not open_temp_tables:
                LOG.info("no open temp tables: %r", open_temp_tables)
                break

            if start_time + self._safe_slave_timeout > time.time():
                LOG.info("time exceeded for open temp tables")
                return False

            prev_slave_coordinates = curr_slave_coordinates

            self._dbcursor.query("START SLAVE SQL_THREAD")
            time.sleep(DB_OPEN_TEMP_TABLES_WAIT)

            curr_slave_coordinates = self._get_slave_coordinates()
            if prev_slave_coordinates \
               and curr_slave_coordinates \
               and prev_slave_coordinates == curr_slave_coordinates:
                LOG.info("Slave pos hasn't moved during wait period, "
                         "not stopping the SQL thread.")
            else:
                self._dbcursor.query("STOP SLAVE SQL_THREAD")

        return open_temp_tables

    def _wait_for_safe_slave(self):
        rs         = False
        start_time = time.time()

        LOG.info("wait for safe slave")

        if not self.check_is_slave():
            LOG.info("db is not a slave")
            return None

        try:
            if self._sql_thread_run:
                self._dbcursor.query("STOP SLAVE SQL_THREAD")

            while True:
                if self._killed:
                    raise GeneratorExit()

                self._check_connection()

                open_temp_tables = self._check_slave_open_temp_tables(start_time)
                if open_temp_tables is False:
                    return False

                if open_temp_tables == 0:
                    self._dbcursor.query("STOP SLAVE SQL_THREAD")
                    open_temp_tables = self._get_slave_open_temp_tables()
                    LOG.info("second check for open_temp_tables: %r", open_temp_tables)
                    if open_temp_tables != 0:
                        continue

                    rs = True
                    LOG.info("slave is safe to backup")
                    return True

                return False
        finally:
            if not rs:
                if self._sql_thread_run:
                    self._dbcursor.query("START SLAVE SQL_THREAD")
                else:
                    self._dbcursor.query("STOP SLAVE SQL_THREAD")

        return rs

    def _acquire_lock(self):
        if self._killed:
            raise GeneratorExit()

        if self._db_locked.is_set():
            LOG.info("db locked already set")
            return

        self._check_connection()

        if self._can_read_only:
            self._sql_read_only = self._get_sql_variable('read_only') in ('1', 'ON')

        if self._safe_slave:
            self._wait_for_safe_slave()

        if not self._no_lock:
            if self._get_sql_variable('lock_wait_timeout') is not None:
                self._dbcursor.query("SET SESSION lock_wait_timeout = %d" % DB_LOCK_WAIT_TIMEOUT)

            if self._lock_tables_for_backup():
                self._db_locked.set()
                return

            if self._get_sql_variable('wresp_on') in ('1', 'ON'):
                self._dbcursor.query("SET SESSION wsrep_causal_reads = 0")

            LOG.info("FLUSH TABLES")

            self._dbcursor.query("FLUSH NO_WRITE_TO_BINLOG TABLES")
            self._dbcursor.query("FLUSH TABLES WITH READ LOCK")

            if self._can_read_only and not self._sql_read_only:
                self._dbcursor.query("SET GLOBAL read_only = ON")

            self._dbcursor.query("FLUSH LOGS")

        self._db_locked.set()

    def _alive(self):
        while not self._killed and self._db_locked.is_set():
            if not self._lock_db.is_set():
                return

            self._check_connection()

            self._dbcursor.query("SELECT SLEEP(%s)" % self._sleep_duration)

    def run(self):
        self._lock_db.wait()
        LOG.info("starting db lock")

        try:
            if self._killed:
                raise GeneratorExit()

            self._dbconn        = anysql.connect_by_uri(self.dburi)
            self._dbcursor      = self._dbconn.cursor()

            self._read_only     = self._get_config_read_only()
            self._can_read_only = self._check_read_only()

            self._acquire_lock()
            self._alive()
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping db lock")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            self._lock_db.clear()
            self._stop()
            self._db_locked.clear()

    def terminate(self):
        LOG.info("db lock terminated")
        self._killed = True
        self._lock_db.set()
        self._db_locked.set()
        self._stop()
