# -*- coding: utf-8 -*-
# Copyright (C) 2020 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.quorum"""

import logging

from dwho.classes.modules import DWhoModuleBase, MODULES
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs import network, xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.abstract.db import XtrargonDBBase


LOG             = logging.getLogger('xtrargon_ng.modules.quorum')

DB_SERVERS_SUFFIX = {'a': 'b',
                     'b': 'a'}

DEFAULT_CNF_FILE  = 'quorum.cnf'

xys.add_callback('quorum.host', network.valid_domain)

class QuorumDB(XtrargonDBBase):
    pass

class QuorumModule(DWhoModuleBase):
    MODULE_NAME = 'quorum'

    LOCK = RWLock()

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']


    HAS_REPLICA_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def has_replica(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.HAS_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        params['host'] = params['host'].lower()
        suffix = params['host'][-1]

        if suffix not in DB_SERVERS_SUFFIX:
            raise HttpReqErrJson(415, "invalid host argument: %r" % params['host'])

        host = "%s%s" % (params['host'][:-1], DB_SERVERS_SUFFIX[suffix])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get replica status on %r", host)
            qdb = QuorumDB()

            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = host)
            qdb.create_connection()

            return qdb.get_slave_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(QuorumModule())
    _start()
